from typing import List, Iterable, Tuple, Dict

from randovania.game_description import default_database
from randovania.game_description.item.major_item import MajorItem
from randovania.games.game import RandovaniaGame
from randovania.layout.base_configuration import BaseConfiguration
from randovania.layout.corruption_configuration import CorruptionConfiguration
from randovania.layout.echoes_configuration import LayoutSkyTempleKeyMode, EchoesConfiguration
from randovania.layout.major_item_state import MajorItemState
from randovania.layout.major_items_configuration import MajorItemsConfiguration
from randovania.layout.pickup_model import PickupModelStyle
from randovania.layout.preset import Preset
from randovania.layout.prime_configuration import PrimeConfiguration


def _bool_to_str(b: bool) -> str:
    if b:
        return "Yes"
    else:
        return "No"


_ECHOES_TEMPLATE_STRINGS = {
    "Item Placement": [
        "Trick Level: {trick_level}",
        "Randomization Mode: {randomization_mode}",
        "Random Starting Items: {random_starting_items}",
    ],
    "Items": [
        "Progressive Suit: {progressive_suit}",
        "Progressive Grapple: {progressive_grapple}",
        "Split Beam Ammo: {split_beam_ammo}",
        "Starting Items: {starting_items}",
        "Item Pool: {item_pool}",
    ],
    "Gameplay": [
        "Starting Location: {starting_location}",
        "Translator Gates: {translator_gates}",
        "Elevators: {elevators}",
    ],
    "Game Changes": [
        "Missiles needs Launcher: {missile_launcher_required}",
        "Power Bombs needs Main PBs: {main_pb_required}",
        "Warp to Start: {warp_to_start}",
        "Final bosses included? {include_final_bosses}",
        "Menu Mod included? {menu_mod}",
    ],
    "Difficulty": [
        "Energy Tank: {energy_tank}",
        "1 HP Mode: {dangerous_energy_tank}",
        "Safe Zone: {safe_zone}",
        "Dark Aether Suit Damage: {dark_aether_suit_damage}",
        "Damage Strictness: {damage_strictness}",
        "Pickup Model: {pickup_model}",
    ],
    "Sky Temple Keys": [
        "Target: {target}",
    ],
}
_CORRUPTION_TEMPLATE_STRINGS = {
    "Item Placement": [
        "Trick Level: {trick_level}",
        "Randomization Mode: {randomization_mode}",
        "Random Starting Items: {random_starting_items}",
    ],
    "Items": [
        "Progressive Missile: {progressive_missile}",
        "Progressive Beam: {progressive_beam}",
        "Starting Items: {starting_items}",
        "Item Pool: {item_pool}",
    ],
    "Gameplay": [
        "Starting Location: {starting_location}",
        "Teleporters: {elevators}",
    ],
    "Game Changes": [
        "Missiles needs Launcher: {missile_launcher_required}",
        "Ship Missile needs Launcher: {ship_launcher_required}",
        "Final bosses included? {include_final_bosses}",
    ],
    "Difficulty": [
        "Energy Tank: {energy_tank}",
        "1 HP Mode: {dangerous_energy_tank}",
        "Damage Strictness: {damage_strictness}",
        "Pickup Model: {pickup_model}",
    ],
}

_PRIME_TEMPLATE_STRINGS = {
    "Item Placement": [
        "Trick Level: {trick_level}",
        "Randomization Mode: {randomization_mode}",
        "Random Starting Items: {random_starting_items}",
    ],
    "Items": [
        "Starting Items: {starting_items}",
        "Item Pool: {item_pool}",
    ],
    "Gameplay": [
        "Starting Location: {starting_location}",
        "Elevators: {elevators}",
    ],
    "Game Changes": [
        "Missiles needs Launcher: {missile_launcher_required}",
        "Power Bomb needs Main: {main_pb_required}",
        "Final bosses included? {include_final_bosses}",
    ],
    "Difficulty": [
        "Energy Tank: {energy_tank}",
        "Damage Strictness: {damage_strictness}",
        "Pickup Model: {pickup_model}",
    ],
}
_EXPECTED_ITEMS = {
    RandovaniaGame.PRIME1: {
        "Scan Visor",
        "Power Beam"
    },
    RandovaniaGame.PRIME2: {
        "Scan Visor",
        "Morph Ball",
        "Power Beam",
        "Charge Beam",
    },
    RandovaniaGame.PRIME3: {
        "Scan Visor",
        "Morph Ball",
        "Morph Ball Bombs",
        "Power Beam",
        "Charge Beam",
        "Space Jump Boots",
    },
}
_CUSTOM_ITEMS = {
    "Cannon Ball",
}

PresetDescription = Tuple[str, List[str]]


def has_shuffled_item(configuration: MajorItemsConfiguration, item_name: str) -> bool:
    for item, state in configuration.items_state.items():
        if item.name == item_name:
            return state.num_shuffled_pickups > 0
    return False


def _calculate_starting_items(game: RandovaniaGame, items_state: Dict[MajorItem, MajorItemState]) -> str:
    expected_items = _EXPECTED_ITEMS[game]
    starting_items = []

    for major_item, item_state in items_state.items():
        if major_item.required:
            continue

        count = item_state.num_included_in_starting_items
        if count > 0:
            if major_item.name in expected_items:
                continue
            if count > 1:
                starting_items.append(f"{count} {major_item.name}")
            else:
                starting_items.append(major_item.name)

        elif major_item.name in expected_items:
            starting_items.append(f"No {major_item.name}")

    if starting_items:
        return ", ".join(starting_items)
    else:
        # If an expected item is missing, it's added as "No X". So empty starting_items means it's precisely vanilla
        return "Vanilla"


def _calculate_item_pool(game: RandovaniaGame, configuration: MajorItemsConfiguration) -> str:
    item_pool = []

    unexpected_items = _EXPECTED_ITEMS[game] | _CUSTOM_ITEMS
    if game == RandovaniaGame.PRIME2:
        if has_shuffled_item(configuration, "Progressive Grapple"):
            unexpected_items.add("Grapple Beam")
            unexpected_items.add("Screw Attack")
        else:
            unexpected_items.add("Progressive Grapple")

        if has_shuffled_item(configuration, "Progressive Suit"):
            unexpected_items.add("Dark Suit")
            unexpected_items.add("Light Suit")
        else:
            unexpected_items.add("Progressive Suit")

    elif game == RandovaniaGame.PRIME3:
        if has_shuffled_item(configuration, "Progressive Beam"):
            unexpected_items.add("Plasma Beam")
            unexpected_items.add("Nova Beam")
        else:
            unexpected_items.add("Progressive Beam")

        if has_shuffled_item(configuration, "Progressive Missile"):
            unexpected_items.add("Ice Missile")
            unexpected_items.add("Seeker Missile")
        else:
            unexpected_items.add("Progressive Missile")

    for major_item, item_state in configuration.items_state.items():
        if major_item.required:
            continue

        item_was_expected = major_item.name not in unexpected_items

        if item_state.num_shuffled_pickups > 0 or item_state.include_copy_in_original_location:
            item_in_pool = True
        else:
            item_in_pool = False

        if item_in_pool:
            if not item_was_expected:
                item_pool.append(major_item.name)
        else:
            if item_was_expected and item_state.num_included_in_starting_items == 0:
                item_pool.append(f"No {major_item.name}")

    if item_pool:
        return ", ".join(item_pool)
    else:
        return "Default"


def _format_params_base(configuration: BaseConfiguration) -> dict:
    game_description = default_database.game_description_for(configuration.game)
    major_items = configuration.major_items_configuration

    format_params = {}

    # Item Placement
    random_starting_items = "{} to {}".format(
        major_items.minimum_random_starting_items,
        major_items.maximum_random_starting_items,
    )
    if random_starting_items == "0 to 0":
        random_starting_items = "None"

    format_params["trick_level"] = configuration.trick_level.pretty_description
    format_params["randomization_mode"] = configuration.available_locations.randomization_mode.value
    format_params["random_starting_items"] = random_starting_items

    # Items
    format_params["starting_items"] = _calculate_starting_items(configuration.game, major_items.items_state)
    format_params["item_pool"] = _calculate_item_pool(configuration.game, major_items)

    # Difficulty
    format_params["damage_strictness"] = configuration.damage_strictness.long_name

    pickup_model = configuration.pickup_model_style.long_name
    if configuration.pickup_model_style != PickupModelStyle.ALL_VISIBLE:
        pickup_model += f" ({configuration.pickup_model_data_source.long_name})"
    format_params["pickup_model"] = pickup_model

    # Gameplay
    starting_locations = configuration.starting_location.locations

    if len(starting_locations) == 1:
        area = game_description.world_list.area_by_area_location(next(iter(starting_locations)))
        format_params["starting_location"] = game_description.world_list.area_name(area)
    else:
        format_params["starting_location"] = "{} locations".format(len(starting_locations))

    return format_params


def _echoes_format_params(configuration: EchoesConfiguration) -> dict:
    major_items = configuration.major_items_configuration
    item_database = default_database.item_database_for_game(configuration.game)

    format_params = {}

    # Items
    unified_ammo = configuration.ammo_configuration.items_state[item_database.ammo["Beam Ammo Expansion"]]

    format_params["progressive_suit"] = _bool_to_str(has_shuffled_item(major_items, "Progressive Suit"))
    format_params["progressive_grapple"] = _bool_to_str(has_shuffled_item(major_items, "Progressive Grapple"))
    format_params["split_beam_ammo"] = _bool_to_str(unified_ammo.pickup_count == 0)

    # Difficulty
    if configuration.varia_suit_damage == 6 and configuration.dark_suit_damage == 1.2:
        dark_aether_suit_damage = "Normal"
    else:
        dark_aether_suit_damage = "Custom"

    format_params["energy_tank"] = f"{configuration.energy_per_tank} energy"
    format_params["dangerous_energy_tank"] = _bool_to_str(configuration.dangerous_energy_tank)
    format_params["safe_zone"] = f"{configuration.safe_zone.heal_per_second} energy/s"
    format_params["dark_aether_suit_damage"] = dark_aether_suit_damage

    # Gameplay
    translator_gates = "Custom"
    translator_configurations = [
        (configuration.translator_configuration.with_vanilla_actual(), "Vanilla (Actual)"),
        (configuration.translator_configuration.with_vanilla_colors(), "Vanilla (Colors)"),
        (configuration.translator_configuration.with_full_random(), "Random"),
    ]
    for translator_config, name in translator_configurations:
        if translator_config == configuration.translator_configuration:
            translator_gates = name
            break

    format_params["translator_gates"] = translator_gates
    format_params["elevators"] = configuration.elevators.value
    format_params["hints"] = "Yes"

    # Game Changes
    missile_launcher_required = True
    main_pb_required = True
    for ammo, state in configuration.ammo_configuration.items_state.items():
        if ammo.name == "Missile Expansion":
            missile_launcher_required = state.requires_major_item
        elif ammo.name == "Power Bomb Expansion":
            main_pb_required = state.requires_major_item

    format_params["missile_launcher_required"] = _bool_to_str(missile_launcher_required)
    format_params["main_pb_required"] = _bool_to_str(main_pb_required)
    format_params["warp_to_start"] = _bool_to_str(configuration.warp_to_start)
    format_params["generic_patches"] = "Some"
    format_params["menu_mod"] = _bool_to_str(configuration.menu_mod)
    format_params["include_final_bosses"] = _bool_to_str(not configuration.skip_final_bosses)

    # Sky Temple Keys
    if configuration.sky_temple_keys.num_keys == LayoutSkyTempleKeyMode.ALL_BOSSES:
        stk_location = "Bosses"
    elif configuration.sky_temple_keys.num_keys == LayoutSkyTempleKeyMode.ALL_GUARDIANS:
        stk_location = "Guardians"
    else:
        stk_location = "Random"

    format_params["target"] = "{0} of {0}".format(configuration.sky_temple_keys.num_keys)
    format_params["location"] = stk_location

    return format_params


def _corruption_format_params(configuration: CorruptionConfiguration) -> dict:
    major_items = configuration.major_items_configuration

    format_params = {"energy_tank": f"{configuration.energy_per_tank} energy",
                     "dangerous_energy_tank": _bool_to_str(configuration.dangerous_energy_tank),
                     "include_final_bosses": _bool_to_str(not configuration.skip_final_bosses),
                     "elevators": configuration.elevators.value,
                     "progressive_missile": _bool_to_str(has_shuffled_item(major_items, "Progressive Missile")),
                     "progressive_beam": _bool_to_str(has_shuffled_item(major_items, "Progressive Beam"))}

    missile_launcher_required = True
    ship_launcher_required = True
    for ammo, state in configuration.ammo_configuration.items_state.items():
        if ammo.name == "Missile Expansion":
            missile_launcher_required = state.requires_major_item
        elif ammo.name == "Ship Missile Expansion":
            ship_launcher_required = state.requires_major_item

    format_params["missile_launcher_required"] = _bool_to_str(missile_launcher_required)
    format_params["ship_launcher_required"] = _bool_to_str(ship_launcher_required)

    return format_params


def _prime_format_params(configuration: PrimeConfiguration) -> dict:
    major_items = configuration.major_items_configuration

    format_params = {"energy_tank": f"{configuration.energy_per_tank} energy",
                     "dangerous_energy_tank": _bool_to_str(configuration.dangerous_energy_tank),
                     "include_final_bosses": _bool_to_str(not configuration.skip_final_bosses),
                     "elevators": configuration.elevators.value
                     }

    missile_launcher_required = True
    main_pb_required = True
    for ammo, state in configuration.ammo_configuration.items_state.items():
        if ammo.name == "Missile Expansion":
            missile_launcher_required = state.requires_major_item
        elif ammo.name == "Power Bomb Expansion":
            main_pb_required = state.requires_major_item

    format_params["missile_launcher_required"] = _bool_to_str(missile_launcher_required)
    format_params["main_pb_required"] = _bool_to_str(main_pb_required)

    return format_params


def describe(preset: Preset) -> Iterable[PresetDescription]:
    configuration = preset.configuration

    format_params = _format_params_base(configuration)

    template_strings = None
    if preset.game == RandovaniaGame.PRIME2:
        template_strings = _ECHOES_TEMPLATE_STRINGS
        format_params.update(_echoes_format_params(configuration))

    elif preset.game == RandovaniaGame.PRIME3:
        template_strings = _CORRUPTION_TEMPLATE_STRINGS
        format_params.update(_corruption_format_params(configuration))

    elif preset.game == RandovaniaGame.PRIME1:
        template_strings = _PRIME_TEMPLATE_STRINGS
        format_params.update(_prime_format_params(configuration))

    for category, templates in template_strings.items():
        yield category, [
            item.format(**format_params)
            for item in templates
        ]


def merge_categories(categories: Iterable[PresetDescription]) -> str:
    return "".join(
        """<h4><span style="font-weight:600;">{}</span></h4><p>{}</p>""".format(category, "<br />".join(items))
        for category, items in categories
    )



import dataclasses
from typing import Optional, Callable, Union

from randovania.games.game import RandovaniaGame
from randovania.layout.ammo_configuration import AmmoConfiguration
from randovania.layout.available_locations import AvailableLocationsConfiguration
from randovania.layout.corruption_configuration import CorruptionConfiguration
from randovania.layout.damage_strictness import LayoutDamageStrictness
from randovania.layout.echoes_configuration import EchoesConfiguration, LayoutSkyTempleKeyMode
from randovania.layout.elevators import LayoutElevators
from randovania.layout.major_items_configuration import MajorItemsConfiguration
from randovania.layout.preset import Preset


class PresetEditor:
    _on_changed: Optional[Callable[[], None]] = None
    _nested_autosave_level: int = 0
    _is_dirty: bool = False

    _name: str
    _base_preset_name: str
    _game: RandovaniaGame
    _configuration: Union[EchoesConfiguration, CorruptionConfiguration]

    def __init__(self, initial_preset: Preset):
        if initial_preset.base_preset_name is None:
            self._name = "{} Custom".format(initial_preset.name)
            self._base_preset_name = initial_preset.name
        else:
            self._name = initial_preset.name
            self._base_preset_name = initial_preset.base_preset_name
        self._game = initial_preset.game
        self._configuration = initial_preset.configuration

    def _set_field(self, field_name: str, value):
        setattr(self, "_" + field_name, value)

    def __enter__(self):
        self._nested_autosave_level += 1
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._nested_autosave_level == 1:
            if self._is_dirty:
                # TODO: maybe it should be an error to change options to different values in on_options_changed?
                if self._on_changed is not None:
                    self._on_changed()
        self._nested_autosave_level -= 1

    # Events
    def _set_on_changed(self, value):
        self._on_changed = value

    on_changed = property(fset=_set_on_changed)

    # Preset

    def create_custom_preset_with(self) -> Preset:
        return Preset(
            name=self.name,
            description="A preset that was customized.",
            base_preset_name=self._base_preset_name,
            game=self._game,
            configuration=self.configuration,
        )

    # Access to Direct fields
    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value):
        self._set_field("name", value)

    @property
    def game(self):
        return self._game

    @property
    def configuration(self) -> Union[EchoesConfiguration, CorruptionConfiguration]:
        return self._configuration

    # Access to fields inside PatcherConfiguration
    @property
    def include_menu_mod(self) -> bool:
        return self.configuration.menu_mod

    @include_menu_mod.setter
    def include_menu_mod(self, value: bool):
        self.set_configuration_field("menu_mod", value)

    @property
    def warp_to_start(self) -> bool:
        return self.configuration.warp_to_start

    @warp_to_start.setter
    def warp_to_start(self, value: bool):
        self.set_configuration_field("warp_to_start", value)

    @property
    def pickup_model_style(self):
        return self.configuration.pickup_model_style

    @pickup_model_style.setter
    def pickup_model_style(self, value):
        self.set_configuration_field("pickup_model_style", value)

    @property
    def pickup_model_data_source(self):
        return self.configuration.pickup_model_data_source

    @pickup_model_data_source.setter
    def pickup_model_data_source(self, value):
        self.set_configuration_field("pickup_model_data_source", value)

    @property
    def layout_configuration_sky_temple_keys(self) -> LayoutSkyTempleKeyMode:
        return self.configuration.sky_temple_keys

    @layout_configuration_sky_temple_keys.setter
    def layout_configuration_sky_temple_keys(self, value: LayoutSkyTempleKeyMode):
        self.set_configuration_field("sky_temple_keys", value)

    @property
    def layout_configuration_damage_strictness(self) -> LayoutDamageStrictness:
        return self.configuration.damage_strictness

    @layout_configuration_damage_strictness.setter
    def layout_configuration_damage_strictness(self, value: LayoutDamageStrictness):
        self.set_configuration_field("damage_strictness", value)

    @property
    def layout_configuration_elevators(self) -> LayoutElevators:
        return self.configuration.elevators

    @layout_configuration_elevators.setter
    def layout_configuration_elevators(self, value: LayoutElevators):
        self.set_configuration_field("elevators", value)

    @property
    def available_locations(self) -> AvailableLocationsConfiguration:
        return self.configuration.available_locations

    @available_locations.setter
    def available_locations(self, value: AvailableLocationsConfiguration):
        self.set_configuration_field("available_locations", value)

    @property
    def major_items_configuration(self) -> MajorItemsConfiguration:
        return self.configuration.major_items_configuration

    @major_items_configuration.setter
    def major_items_configuration(self, value: MajorItemsConfiguration):
        self.set_configuration_field("major_items_configuration", value)

    @property
    def ammo_configuration(self) -> AmmoConfiguration:
        return self.configuration.ammo_configuration

    @ammo_configuration.setter
    def ammo_configuration(self, value: AmmoConfiguration):
        self.set_configuration_field("ammo_configuration", value)

    @property
    def skip_final_bosses(self) -> bool:
        return self.configuration.skip_final_bosses

    @skip_final_bosses.setter
    def skip_final_bosses(self, value: bool):
        self.set_configuration_field("skip_final_bosses", value)

    def set_configuration_field(self, field_name: str, value):
        self._edit_field(
            "configuration",
            dataclasses.replace(self.configuration, **{field_name: value})
        )

    ######

    def _check_editable_and_mark_dirty(self):
        """Checks if _nested_autosave_level is not 0 and marks at least one value was changed."""
        assert self._nested_autosave_level != 0, "Attempting to edit an PresetEditor, but it wasn't made editable"
        self._is_dirty = True

    def _edit_field(self, field_name: str, new_value):
        current_value = getattr(self, field_name)
        if current_value != new_value:
            self._check_editable_and_mark_dirty()
            self._set_field(field_name, new_value)



from randovania.games.game import RandovaniaGame
from randovania.games.prime.all_prime_dol_patches import StringDisplayPatchAddresses, HealthCapacityAddresses, \
    DangerousEnergyTankAddresses
from randovania.games.prime.corruption_dol_patches import CorruptionDolVersion
from randovania.games.prime.echoes_dol_patches import BeamCostAddresses, SafeZoneAddresses, EchoesDolVersion

ALL_VERSIONS = [
    CorruptionDolVersion(
        game=RandovaniaGame.PRIME3,
        description="Wii NTSC",
        build_string_address=0x805822b0,
        build_string=b"!#$MetroidBuildInfo!#$2007/07/27 13:13 Build v3.436 (MP3)",
        sda2_base=0x806869C0,
        game_state_pointer=None,
        string_display=StringDisplayPatchAddresses(
            update_hint_state=None,
            message_receiver_string_ref=0x805a4200,
            wstring_constructor=None,
            display_hud_memo=None,
            cstate_manager_global=0x8067dc0c,
            max_message_size=200,
        ),
        health_capacity=HealthCapacityAddresses(
            base_health_capacity=0x80680a6c,
            energy_tank_capacity=0x80680a70,
        ),
        dangerous_energy_tank=DangerousEnergyTankAddresses(
            small_number_float=0x8067fe8c,
            incr_pickup=0x801913a0,
        ),
        game_options_constructor_address=None,
    ),
]



from typing import List, Dict, NamedTuple

from randovania.game_description.resources.pickup_entry import PickupEntry
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.resources.resource_info import CurrentResources


class PoolResults(NamedTuple):
    pickups: List[PickupEntry]
    assignment: Dict[PickupIndex, PickupEntry]
    initial_resources: CurrentResources



import dataclasses
from typing import NamedTuple, List, Iterator, Optional, Dict

from randovania.game_description.area import Area
from randovania.game_description.node import Node
from randovania.game_description.resources.pickup_index import PickupIndex


@dataclasses.dataclass(frozen=True)
class World:
    name: str
    dark_name: Optional[str]
    world_asset_id: int
    areas: List[Area]

    def __post_init__(self):
        object.__setattr__(self, "__cached_area_by_asset_id", {})

    def __repr__(self):
        return "World[{}]".format(self.name)

    @property
    def all_nodes(self) -> Iterator[Node]:
        """
        Iterates over all nodes in all areas of this world.
        :return:
        """
        for area in self.areas:
            yield from area.nodes

    @property
    def pickup_indices(self) -> Iterator[PickupIndex]:
        for area in self.areas:
            yield from area.pickup_indices

    @property
    def major_pickup_indices(self) -> Iterator[PickupIndex]:
        for area in self.areas:
            yield from area.major_pickup_indices

    def area_by_asset_id(self, asset_id: int) -> Area:
        cache: Dict[int, int] = object.__getattribute__(self, "__cached_area_by_asset_id")
        if asset_id in cache:
            return self.areas[cache[asset_id]]

        for i, area in enumerate(self.areas):
            if area.area_asset_id == asset_id:
                cache[asset_id] = i
                return area
        raise KeyError("Unknown asset_id: {}".format(asset_id))

    def area_by_name(self, area_name: str) -> Area:
        for area in self.areas:
            if area.name == area_name:
                return area
        raise KeyError("Unknown name: {}".format(area_name))

    def correct_name(self, in_dark_world: bool) -> str:
        if in_dark_world and self.dark_name is not None:
            return self.dark_name
        return self.name



from typing import List, Dict, Optional, Any, Tuple

from PySide2.QtCore import QTimer, Signal, QObject
from asyncqt import asyncSlot

from randovania.game_connection.connection_backend import ConnectionBackend
from randovania.game_connection.connection_base import GameConnectionStatus, ConnectionBase, InventoryItem
from randovania.game_description.resources.item_resource_info import ItemResourceInfo
from randovania.game_description.resources.pickup_entry import PickupEntry


class GameConnection(QObject, ConnectionBase):
    Updated = Signal()

    _dt: float = 2.5
    _last_status: Any = None
    backend: ConnectionBackend
    _permanent_pickups: List[Tuple[str, PickupEntry]]

    def __init__(self, backend: ConnectionBackend):
        super().__init__()
        self._permanent_pickups = []

        self._timer = QTimer(self)
        self._timer.timeout.connect(self._update)
        self._timer.setInterval(self._dt * 1000)
        self._timer.setSingleShot(True)

        self.set_backend(backend)

    def set_backend(self, backend: ConnectionBackend):
        if hasattr(self, "backend"):
            self.backend.set_location_collected_listener(None)
        self.backend = backend
        self.backend.set_location_collected_listener(self._emit_location_collected)
        self.backend.checking_for_collected_index = self._location_collected_listener is not None
        self.backend.set_permanent_pickups(self._permanent_pickups)
        self._notify_status()

    async def start(self):
        self._timer.start()

    async def stop(self):
        self._timer.stop()

    @asyncSlot()
    async def _update(self):
        try:
            await self.backend.update(self._dt)
            self._notify_status()
        finally:
            self._timer.start()

    def _notify_status(self):
        new_status = self.current_status
        if self._last_status != (new_status, self.backend):
            self._last_status = (new_status, self.backend)
            self.Updated.emit()

    @property
    def pretty_current_status(self) -> str:
        return f"{self.backend.name}: {self.backend.current_status.pretty_text}"

    @property
    def current_status(self) -> GameConnectionStatus:
        return self.backend.current_status

    def get_current_inventory(self) -> Dict[ItemResourceInfo, InventoryItem]:
        return self.backend.get_current_inventory()

    def set_permanent_pickups(self, pickups: List[Tuple[str, PickupEntry]]):
        self._permanent_pickups = pickups
        return self.backend.set_permanent_pickups(pickups)

    def set_location_collected_listener(self, listener):
        super().set_location_collected_listener(listener)
        self.backend.checking_for_collected_index = listener is not None




import re
from typing import Optional

from PySide2.QtWidgets import QDialog, QWidget

from randovania.game_description.area import Area
from randovania.game_description.game_description import GameDescription
from randovania.game_description.node import DockNode
from randovania.game_description.requirements import RequirementList, Requirement
from randovania.game_description.resources.resource_type import ResourceType
from randovania.game_description.resources.trick_resource_info import TrickResourceInfo
from randovania.gui.generated.trick_details_popup_ui import Ui_TrickDetailsPopup
from randovania.gui.lib.common_qt_lib import set_default_window_icon
from randovania.gui.lib.window_manager import WindowManager
from randovania.layout.trick_level import LayoutTrickLevel


def _has_trick(alternative: RequirementList) -> bool:
    return any(individual.resource.resource_type == ResourceType.TRICK for individual in alternative.values())


def _area_uses_trick(area: Area,
                     trick: TrickResourceInfo,
                     level: LayoutTrickLevel,
                     ) -> bool:
    """
    Checks the area RequirementSet in the given Area uses the given trick at the given level.
    :param area:
    :param trick:
    :param level:
    :return:
    """

    def _uses_trick(requirements: Requirement) -> bool:
        return any(individual.resource == trick and individual.amount == level.as_number
                   for individual in requirements.as_set.all_individual)

    for node in area.nodes:
        if isinstance(node, DockNode):
            if _uses_trick(node.default_dock_weakness.requirement):
                return True

        if any(_uses_trick(req) for req in area.connections[node].values()):
            return True

    return False


class TrickDetailsPopup(QDialog, Ui_TrickDetailsPopup):
    def __init__(self,
                 parent: QWidget,
                 window_manager: WindowManager,
                 game_description: GameDescription,
                 trick: TrickResourceInfo,
                 level: LayoutTrickLevel,
                 ):
        super().__init__(parent)
        self.setupUi(self)
        set_default_window_icon(self)

        self._window_manager = window_manager
        self._game_description = game_description

        # setup
        self.setWindowTitle(f"Trick Details: {trick.long_name} at {level.long_name}")
        self.title_label.setText(self.title_label.text().format(
            trick=trick,
            level=level.long_name,
        ))
        self.area_list_label.linkActivated.connect(self._on_click_link_to_data_editor)

        # connect
        self.button_box.accepted.connect(self.button_box_close)
        self.button_box.rejected.connect(self.button_box_close)

        # Update
        areas_to_show = [
            (world, area)
            for world in game_description.world_list.worlds
            for area in world.areas
            if _area_uses_trick(area, trick, level)
        ]

        if areas_to_show:
            lines = [
                f'<a href="data-editor://{world.correct_name(area.in_dark_aether)}/{area.name}">{world.correct_name(area.in_dark_aether)} - {area.name}</a>'
                for (world, area) in areas_to_show
            ]
            self.area_list_label.setText("<br />".join(sorted(lines)))

        else:
            self.area_list_label.setText("This trick is not used in this level.")

    def button_box_close(self):
        self.reject()

    def _on_click_link_to_data_editor(self, link: str):
        info = re.match(r"^data-editor://([^)]+)/([^)]+)$", link)
        if info:
            world_name, area_name = info.group(1, 2)
            self._window_manager.open_data_visualizer_at(world_name, area_name, game=self._game_description.game)



from unittest.mock import MagicMock, call

import flask
import pytest

from randovania.network_common.error import NotLoggedIn, ServerError, InvalidSession
from randovania.server import database
from randovania.server.server_app import ServerApp


@pytest.fixture(name="server_app")
def server_app_fixture(flask_app, skip_qtbot):
    flask_app.config['SECRET_KEY'] = "key"
    flask_app.config["DISCORD_CLIENT_ID"] = 1234
    flask_app.config["DISCORD_CLIENT_SECRET"] = 5678
    flask_app.config["DISCORD_REDIRECT_URI"] = "http://127.0.0.1:5000/callback/"
    flask_app.config["FERNET_KEY"] = b's2D-pjBIXqEqkbeRvkapeDn82MgZXLLQGZLTgqqZ--A='
    flask_app.config["GUEST_KEY"] = b's2D-pjBIXqEqkbeRvkapeDn82MgZXLLQGZLTgqqZ--A='
    server = ServerApp(flask_app)
    server.metrics.summary = MagicMock()
    server.metrics.summary.return_value.side_effect = lambda x: x
    return server


def test_session(server_app):
    server_app.sio = MagicMock()

    with server_app.app.test_request_context():
        flask.request.sid = 1234
        result = server_app.session()

    assert result == server_app.sio.server.session.return_value
    server_app.sio.server.session.assert_called_once_with(1234, namespace=None)


def test_get_session(server_app):
    server_app.sio = MagicMock()

    with server_app.app.test_request_context():
        flask.request.sid = 1234
        result = server_app.get_session()

    assert result == server_app.sio.server.get_session.return_value
    server_app.sio.server.get_session.assert_called_once_with(1234, namespace=None)


def test_get_current_user_ok(server_app, clean_database):
    server_app.get_session = MagicMock(return_value={"user-id": 1234})
    user = database.User.create(id=1234, name="Someone")

    # Run
    result = server_app.get_current_user()

    # Assert
    assert result == user


def test_get_current_user_not_logged(server_app, clean_database):
    server_app.get_session = MagicMock(return_value={})

    # Run
    with pytest.raises(NotLoggedIn):
        server_app.get_current_user()


def test_get_current_user_unknown_user(server_app, clean_database):
    server_app.get_session = MagicMock(return_value={"user-id": 1234})

    # Run
    with pytest.raises(InvalidSession):
        server_app.get_current_user()


def test_join_game_session(mocker, server_app):
    mock_join_room = mocker.patch("flask_socketio.join_room")
    membership = MagicMock()
    membership.session.id = "session_id"
    membership.user.id = "user_id"

    session = {}
    server_app.session = MagicMock()
    server_app.session.return_value.__enter__.return_value = session

    # Run
    server_app.join_game_session(membership)

    # Assert
    mock_join_room.assert_has_calls([
        call("game-session-session_id"),
        call("game-session-session_id-user_id"),
    ])
    assert session == {
        "current_game_session": "session_id",
    }


def test_leave_game_session_with_session(mocker, server_app):
    # Setup
    mock_leave_room = mocker.patch("flask_socketio.leave_room")
    user = MagicMock()
    user.id = "user_id"
    server_app.get_current_user = lambda: user

    session = {"current_game_session": "session_id"}
    server_app.session = MagicMock()
    server_app.session.return_value.__enter__.return_value = session

    # Run
    server_app.leave_game_session()

    # Assert
    mock_leave_room.assert_has_calls([
        call("game-session-session_id"),
        call("game-session-session_id-user_id"),
    ])
    assert session == {}


def test_leave_game_session_without_session(mocker, server_app):
    # Setup
    mock_leave_room: MagicMock = mocker.patch("flask_socketio.leave_room")
    server_app.session = MagicMock()
    server_app.session.return_value.__enter__.return_value = {}

    # Run
    server_app.leave_game_session()

    # Assert
    mock_leave_room.assert_not_called()


def test_on_success_ok(server_app):
    # Setup
    custom = MagicMock(return_value={"foo": 12345})
    server_app.on("custom", custom)

    # Run
    test_client = server_app.sio.test_client(server_app.app)
    result = test_client.emit("custom", callback=True)

    # Assert
    custom.assert_called_once_with(server_app)
    assert result == {"result": {"foo": 12345}}


def test_on_success_network_error(server_app):
    # Setup
    error = NotLoggedIn()
    custom = MagicMock(side_effect=error)
    server_app.on("custom", custom)

    # Run
    test_client = server_app.sio.test_client(server_app.app)
    result = test_client.emit("custom", callback=True)

    # Assert
    custom.assert_called_once_with(server_app)
    assert result == error.as_json


def test_on_success_exception(server_app):
    # Setup
    custom = MagicMock(side_effect=RuntimeError("something happened"))
    server_app.on("custom", custom)

    # Run
    test_client = server_app.sio.test_client(server_app.app)
    result = test_client.emit("custom", callback=True)

    # Assert
    custom.assert_called_once_with(server_app)
    assert result == ServerError().as_json



import itertools

import pytest

from randovania.bitpacking import bitpacking
from randovania.bitpacking.bitpacking import BitPackDecoder
from randovania.game_description.area_location import AreaLocation
from randovania.games.game import RandovaniaGame
from randovania.layout.starting_location import StartingLocation


@pytest.fixture(
    params=[
        {"encoded": b'\x00', "json": []},
        {"encoded": b'\x0c\x00', "json": ["Temple Grounds/Landing Site"]},
        {"encoded": b'\x14\x84', "json": ["Temple Grounds/Hall of Honored Dead", "Temple Grounds/Path of Eyes"]},
    ],
    name="location_with_data")
def _location_with_data(request, mocker, echoes_game_description):
    world_list = echoes_game_description.world_list
    areas = list(itertools.islice(
        (AreaLocation(world.world_asset_id, area.area_asset_id)
         for world in world_list.worlds
         for area in world.areas
         if area.valid_starting_location), 15))

    mocker.patch("randovania.layout.starting_location._areas_list",
                 return_value=list(sorted(areas)))
    return request.param["encoded"], StartingLocation.from_json(request.param["json"], RandovaniaGame.PRIME2)


def test_decode(location_with_data):
    # Setup
    data, expected = location_with_data

    # Run
    decoder = BitPackDecoder(data)
    result = StartingLocation.bit_pack_unpack(
        decoder, {"reference": StartingLocation.with_elements([], RandovaniaGame.PRIME2)})

    # Assert
    assert result == expected


def test_encode(location_with_data):
    # Setup
    expected, value = location_with_data

    # Run
    result = bitpacking.pack_value(value)

    # Assert
    assert result == expected



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'data_editor.ui'
##
## Created by: Qt User Interface Compiler version 5.14.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import (QCoreApplication, QDate, QDateTime, QMetaObject,
    QObject, QPoint, QRect, QSize, QTime, QUrl, Qt)
from PySide2.QtGui import (QBrush, QColor, QConicalGradient, QCursor, QFont,
    QFontDatabase, QIcon, QKeySequence, QLinearGradient, QPalette, QPainter,
    QPixmap, QRadialGradient)
from PySide2.QtWidgets import *


class Ui_DataEditorWindow(object):
    def setupUi(self, DataEditorWindow):
        if not DataEditorWindow.objectName():
            DataEditorWindow.setObjectName(u"DataEditorWindow")
        DataEditorWindow.resize(802, 471)
        self.centralWidget = QWidget(DataEditorWindow)
        self.centralWidget.setObjectName(u"centralWidget")
        self.centralWidget.setMaximumSize(QSize(16777215, 16777215))
        self.gridLayout_2 = QGridLayout(self.centralWidget)
        self.gridLayout_2.setSpacing(6)
        self.gridLayout_2.setContentsMargins(11, 11, 11, 11)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.points_of_interest_group = QGroupBox(self.centralWidget)
        self.points_of_interest_group.setObjectName(u"points_of_interest_group")
        self.points_of_interest_group.setMaximumSize(QSize(250, 16777215))
        self.verticalLayout = QVBoxLayout(self.points_of_interest_group)
        self.verticalLayout.setSpacing(6)
        self.verticalLayout.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.delete_node_button = QPushButton(self.points_of_interest_group)
        self.delete_node_button.setObjectName(u"delete_node_button")

        self.verticalLayout.addWidget(self.delete_node_button)

        self.new_node_button = QPushButton(self.points_of_interest_group)
        self.new_node_button.setObjectName(u"new_node_button")

        self.verticalLayout.addWidget(self.new_node_button)


        self.gridLayout_2.addWidget(self.points_of_interest_group, 1, 0, 2, 1)

        self.save_database_button = QPushButton(self.centralWidget)
        self.save_database_button.setObjectName(u"save_database_button")

        self.gridLayout_2.addWidget(self.save_database_button, 3, 0, 1, 1)

        self.connections_group = QGroupBox(self.centralWidget)
        self.connections_group.setObjectName(u"connections_group")
        self.connections_group.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)
        self.gridLayout = QGridLayout(self.connections_group)
        self.gridLayout.setSpacing(6)
        self.gridLayout.setContentsMargins(11, 11, 11, 11)
        self.gridLayout.setObjectName(u"gridLayout")
        self.other_node_connection_combo = QComboBox(self.connections_group)
        self.other_node_connection_combo.setObjectName(u"other_node_connection_combo")
        self.other_node_connection_combo.setEnabled(False)

        self.gridLayout.addWidget(self.other_node_connection_combo, 0, 0, 1, 1)

        self.other_node_connection_edit_button = QPushButton(self.connections_group)
        self.other_node_connection_edit_button.setObjectName(u"other_node_connection_edit_button")

        self.gridLayout.addWidget(self.other_node_connection_edit_button, 0, 1, 1, 1)

        self.other_node_alternatives_scroll_area = QScrollArea(self.connections_group)
        self.other_node_alternatives_scroll_area.setObjectName(u"other_node_alternatives_scroll_area")
        self.other_node_alternatives_scroll_area.setLineWidth(4)
        self.other_node_alternatives_scroll_area.setWidgetResizable(True)
        self.other_node_alternatives_contents = QWidget()
        self.other_node_alternatives_contents.setObjectName(u"other_node_alternatives_contents")
        self.other_node_alternatives_contents.setGeometry(QRect(0, 0, 487, 140))
        self.other_node_alternatives_scroll_area.setWidget(self.other_node_alternatives_contents)

        self.gridLayout.addWidget(self.other_node_alternatives_scroll_area, 1, 0, 1, 2)


        self.gridLayout_2.addWidget(self.connections_group, 2, 1, 2, 1)

        self.world_selection_group = QGroupBox(self.centralWidget)
        self.world_selection_group.setObjectName(u"world_selection_group")
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.world_selection_group.sizePolicy().hasHeightForWidth())
        self.world_selection_group.setSizePolicy(sizePolicy)
        self.horizontalLayout = QHBoxLayout(self.world_selection_group)
        self.horizontalLayout.setSpacing(6)
        self.horizontalLayout.setContentsMargins(11, 11, 11, 11)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.world_selector_box = QComboBox(self.world_selection_group)
        self.world_selector_box.setObjectName(u"world_selector_box")

        self.horizontalLayout.addWidget(self.world_selector_box)

        self.area_selector_box = QComboBox(self.world_selection_group)
        self.area_selector_box.setObjectName(u"area_selector_box")
        self.area_selector_box.setEnabled(False)

        self.horizontalLayout.addWidget(self.area_selector_box)


        self.gridLayout_2.addWidget(self.world_selection_group, 0, 0, 1, 2)

        self.node_info_group = QGroupBox(self.centralWidget)
        self.node_info_group.setObjectName(u"node_info_group")
        sizePolicy.setHeightForWidth(self.node_info_group.sizePolicy().hasHeightForWidth())
        self.node_info_group.setSizePolicy(sizePolicy)
        self.gridLayout_4 = QGridLayout(self.node_info_group)
        self.gridLayout_4.setSpacing(6)
        self.gridLayout_4.setContentsMargins(11, 11, 11, 11)
        self.gridLayout_4.setObjectName(u"gridLayout_4")
        self.node_name_label = QLabel(self.node_info_group)
        self.node_name_label.setObjectName(u"node_name_label")

        self.gridLayout_4.addWidget(self.node_name_label, 0, 0, 1, 1)

        self.node_heals_check = QCheckBox(self.node_info_group)
        self.node_heals_check.setObjectName(u"node_heals_check")
        self.node_heals_check.setMaximumSize(QSize(100, 16777215))

        self.gridLayout_4.addWidget(self.node_heals_check, 0, 1, 1, 1)

        self.node_edit_button = QPushButton(self.node_info_group)
        self.node_edit_button.setObjectName(u"node_edit_button")
        self.node_edit_button.setMaximumSize(QSize(100, 16777215))

        self.gridLayout_4.addWidget(self.node_edit_button, 0, 2, 1, 1)

        self.node_details_label = QLabel(self.node_info_group)
        self.node_details_label.setObjectName(u"node_details_label")

        self.gridLayout_4.addWidget(self.node_details_label, 1, 0, 1, 2)

        self.area_spawn_check = QCheckBox(self.node_info_group)
        self.area_spawn_check.setObjectName(u"area_spawn_check")

        self.gridLayout_4.addWidget(self.area_spawn_check, 1, 2, 1, 1)


        self.gridLayout_2.addWidget(self.node_info_group, 1, 1, 1, 1)

        DataEditorWindow.setCentralWidget(self.centralWidget)

        self.retranslateUi(DataEditorWindow)

        QMetaObject.connectSlotsByName(DataEditorWindow)
    # setupUi

    def retranslateUi(self, DataEditorWindow):
        DataEditorWindow.setWindowTitle(QCoreApplication.translate("DataEditorWindow", u"Data Visualizer", None))
        self.points_of_interest_group.setTitle(QCoreApplication.translate("DataEditorWindow", u"Nodes", None))
        self.delete_node_button.setText(QCoreApplication.translate("DataEditorWindow", u"Delete Node", None))
        self.new_node_button.setText(QCoreApplication.translate("DataEditorWindow", u"New Node", None))
        self.save_database_button.setText(QCoreApplication.translate("DataEditorWindow", u"Save as database", None))
        self.connections_group.setTitle(QCoreApplication.translate("DataEditorWindow", u"Connections", None))
        self.other_node_connection_edit_button.setText(QCoreApplication.translate("DataEditorWindow", u"Edit", None))
        self.world_selection_group.setTitle(QCoreApplication.translate("DataEditorWindow", u"World/Area Selection", None))
        self.node_info_group.setTitle(QCoreApplication.translate("DataEditorWindow", u"Node Info", None))
        self.node_name_label.setText(QCoreApplication.translate("DataEditorWindow", u"TextLabel", None))
        self.node_heals_check.setText(QCoreApplication.translate("DataEditorWindow", u"Heals?", None))
        self.node_edit_button.setText(QCoreApplication.translate("DataEditorWindow", u"Edit", None))
        self.node_details_label.setText(QCoreApplication.translate("DataEditorWindow", u"TextLabel", None))
        self.area_spawn_check.setText(QCoreApplication.translate("DataEditorWindow", u"Area spawn", None))
    # retranslateUi




import collections
import typing
from random import Random
from typing import Dict, Tuple, Optional

from randovania.game_description import node_search
from randovania.game_description.area import Area
from randovania.game_description.assignment import PickupAssignment, PickupTarget
from randovania.game_description.game_patches import GamePatches
from randovania.game_description.hint import HintType, HintLocationPrecision, HintItemPrecision, Hint, RelativeDataArea, \
    HintRelativeAreaName, RelativeDataItem, HintDarkTemple
from randovania.game_description.item.item_category import ItemCategory
from randovania.game_description.node import LogbookNode, PickupNode
from randovania.game_description.resources import resource_info
from randovania.game_description.resources.pickup_entry import PickupEntry, ConditionalResources
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.world_list import WorldList
from randovania.games.prime import echoes_items
from randovania.games.prime.echoes_items import USELESS_PICKUP_MODEL
from randovania.games.prime.patcher_file_lib.hint_name_creator import LocationHintCreator, create_simple_logbook_hint, \
    color_text, TextColor
from randovania.interface_common.players_configuration import PlayersConfiguration

_JOKE_HINTS = [
    # Guidelines for joke hints:
    # 1. They should clearly be jokes, and not real hints or the result of a bug.
    # 2. They shouldn't reference real-world people.
    # 3. They should be understandable by as many people as possible.
    #
    "By this point in your run, you should have consumed at least 200 mL of water to maintain optimum hydration.",
    "Make sure to collect an Energy Transfer Module; otherwise your run won't be valid!",
    "Adam has not yet authorized the use of this hint.",
    "Back in my day, we didn't need hints!",
    "Hear the words of O-Lir, last Sentinel of the Fortress Temple. May they serve you well.",
    "Warning! Dark Aether's atmosphere is dangerous! Energized Safe Zones don't last forever!",
    "A really important item can be found at - (transmission ends)",
    "Did you know that Bigfoot and Santa Claus exist in the Metroid Prime canon?",
    "I hear. Them. Everywhere. They're coming. Can't sleep. Ever. They'll eat me. Eat.",
    "Space Pirates, strangely, dislike theft.",
    "Power Bomb Expansions are just space hamburgers.",
    "While in Morph Ball mode, press &image=SI,0.70,0.68,D523DE3B; to drop Bombs.",
    "Charge your beam to fire a normal shot when out of ammo.",
    "Movement in Morph Ball mode is faster than unmorphed, even without Boost Ball.",
    "While walking, holding L makes you move faster."
]

_DET_AN = [
    "Annihilator Beam",
    "Amber Translator",
    "Echo Visor",
    "Emerald Translator",
    "Energy Transfer Module",
    "Energy Tank",
]

_DET_NULL = []
_DET_NULL.extend(f"{temple} Key {i}"
                 for i in range(1, 4)
                 for temple in ("Dark Agon", "Dark Torvus", "Ing Hive"))
_DET_NULL.extend(f"Sky Temple Key {i}" for i in range(1, 10))


class Determiner:
    s: str
    supports_title: bool

    def __init__(self, s, supports_title: bool = True):
        self.s = s
        self.supports_title = supports_title

    def __format__(self, format_spec):
        return self.s.__format__(format_spec)

    @property
    def title(self):
        if self.supports_title:
            return self.s.title()
        else:
            return self.s


class LocationFormatter:
    def format(self, determiner: Determiner, pickup_name: str, hint: Hint) -> str:
        raise NotImplementedError()


class GuardianFormatter(LocationFormatter):
    _GUARDIAN_NAMES = {
        PickupIndex(43): "Amorbis",
        PickupIndex(79): "Chykka",
        PickupIndex(115): "Quadraxis",
    }

    def format(self, determiner: Determiner, pickup: str, hint: Hint) -> str:
        guardian = color_text(TextColor.GUARDIAN, self._GUARDIAN_NAMES[hint.target])
        return f"{guardian} is guarding {determiner}{pickup}."


class TemplatedFormatter(LocationFormatter):
    def __init__(self, template: str, hint_name_creator: LocationHintCreator):
        self.template = template
        self.hint_name_creator = hint_name_creator

    def format(self, determiner: Determiner, pickup: str, hint: Hint) -> str:
        node_name = color_text(TextColor.LOCATION, self.hint_name_creator.index_node_name(
            hint.target,
            hint.precision.location == HintLocationPrecision.WORLD_ONLY
        ))
        return self.template.format(determiner=determiner,
                                    pickup=pickup,
                                    node=node_name)


class RelativeFormatter(LocationFormatter):
    def __init__(self, world_list: WorldList, patches: GamePatches):
        self.world_list = world_list
        self.patches = patches
        self._index_to_node = {
            node.pickup_index: node
            for node in world_list.all_nodes
            if isinstance(node, PickupNode)
        }

    def _calculate_distance(self, source_location: PickupIndex, target: Area) -> int:
        source = self._index_to_node[source_location]
        return node_search.distances_to_node(self.world_list, source,
                                             patches=self.patches, ignore_elevators=False)[target]

    def relative_format(self, determiner: Determiner, pickup: str, hint: Hint, other_area: Area, other_name: str,
                        ) -> str:
        distance = self._calculate_distance(hint.target, other_area)
        if distance > 1:
            precise_msg = "exactly " if hint.precision.relative.precise_distance else "up to "
            distance_msg = f"{precise_msg}{distance} rooms"
        else:
            distance_msg = "one room"

        return (f"{determiner.title}{pickup} can be found "
                f"{color_text(TextColor.LOCATION, distance_msg)} away from {other_name}.")


class RelativeAreaFormatter(RelativeFormatter):
    def format(self, determiner: Determiner, pickup: str, hint: Hint) -> str:
        relative = typing.cast(RelativeDataArea, hint.precision.relative)
        other_area = self.world_list.area_by_area_location(relative.area_location)

        if relative.precision == HintRelativeAreaName.NAME:
            other_name = self.world_list.area_name(other_area)
        elif relative.precision == HintRelativeAreaName.FEATURE:
            raise NotImplementedError("HintRelativeAreaName.FEATURE not implemented")
        else:
            raise ValueError(f"Unknown precision: {relative.precision}")

        return self.relative_format(determiner, pickup, hint, other_area, other_name)


class RelativeItemFormatter(RelativeFormatter):
    def __init__(self, world_list: WorldList, patches: GamePatches, players_config: PlayersConfiguration):
        super().__init__(world_list, patches)
        self.players_config = players_config

    def format(self, determiner: Determiner, pickup: str, hint: Hint) -> str:
        relative = typing.cast(RelativeDataItem, hint.precision.relative)
        index = relative.other_index

        other_area = self.world_list.nodes_to_area(self._index_to_node[index])
        other_name = "".join(_calculate_pickup_hint(self.patches.pickup_assignment, self.world_list,
                                                    relative.precision,
                                                    self.patches.pickup_assignment.get(index),
                                                    self.players_config))

        return self.relative_format(determiner, pickup, hint, other_area, other_name)


def _calculate_pickup_hint(pickup_assignment: PickupAssignment,
                           world_list: WorldList,
                           precision: HintItemPrecision,
                           target: Optional[PickupTarget],
                           players_config: PlayersConfiguration,
                           ) -> Tuple[str, str]:
    """

    :param pickup_assignment:
    :param world_list:
    :param precision:
    :param target:
    :return:
    """
    if target is None:
        target = PickupTarget(
            pickup=PickupEntry(
                name="Energy Transfer Module",
                resources=(ConditionalResources(None, None, ()),),
                model_index=USELESS_PICKUP_MODEL,
                item_category=ItemCategory.ETM,
                broad_category=ItemCategory.ETM,
            ),
            player=players_config.player_index,
        )

    if precision is HintItemPrecision.GENERAL_CATEGORY:
        return target.pickup.item_category.general_details

    elif precision is HintItemPrecision.PRECISE_CATEGORY:
        return target.pickup.item_category.hint_details

    elif precision is HintItemPrecision.BROAD_CATEGORY:
        return target.pickup.broad_category.hint_details

    elif precision is HintItemPrecision.DETAILED:
        return _calculate_determiner(pickup_assignment, target.pickup, world_list), target.pickup.name

    elif precision is HintItemPrecision.NOTHING:
        return "an", "item"

    else:
        raise ValueError(f"Unknown precision: {precision}")


def _calculate_determiner(pickup_assignment: PickupAssignment, pickup: PickupEntry, world_list: WorldList) -> str:
    name_count = collections.defaultdict(int)
    for i in range(world_list.num_pickup_nodes):
        index = PickupIndex(i)
        if index in pickup_assignment:
            pickup_name = pickup_assignment[index].pickup.name
        else:
            pickup_name = "Energy Transfer Module"
        name_count[pickup_name] += 1

    if pickup.name in _DET_NULL:
        determiner = ""
    elif name_count[pickup.name] == 1:
        determiner = "the "
    elif pickup.name in _DET_AN:
        determiner = "an "
    else:
        determiner = "a "

    return determiner


def create_temple_key_hint(all_patches: Dict[int, GamePatches],
                           player_index: int,
                           temple: HintDarkTemple,
                           world_list: WorldList,
                           ) -> str:
    """
    Creates the text for .
    :param all_patches:
    :param player_index:
    :param temple:
    :param world_list:
    :return:
    """
    all_world_names = set()

    _TEMPLE_NAMES = ["Dark Agon Temple", "Dark Torvus Temple", "Hive Temple"]
    temple_index = [HintDarkTemple.AGON_WASTES, HintDarkTemple.TORVUS_BOG,
                    HintDarkTemple.SANCTUARY_FORTRESS].index(temple)
    keys = echoes_items.DARK_TEMPLE_KEY_ITEMS[temple_index]

    index_to_node = {
        node.pickup_index: node
        for node in world_list.all_nodes
        if isinstance(node, PickupNode)
    }

    for patches in all_patches.values():
        for pickup_index, target in patches.pickup_assignment.items():
            if target.player != player_index:
                continue

            resources = resource_info.convert_resource_gain_to_current_resources(target.pickup.resource_gain({}))
            for resource, quantity in resources.items():
                if quantity < 1 or resource.index not in keys:
                    continue

                pickup_node = index_to_node[pickup_index]
                all_world_names.add(world_list.world_name_from_node(pickup_node, True))

    temple_name = color_text(TextColor.ITEM, _TEMPLE_NAMES[temple_index])
    names_sorted = [color_text(TextColor.LOCATION, world) for world in sorted(all_world_names)]
    if len(names_sorted) == 0:
        return f"The keys to {temple_name} are nowhere to be found."
    elif len(names_sorted) == 1:
        return f"The keys to {temple_name} can all be found in {names_sorted[0]}."
    else:
        last = names_sorted.pop()
        front = ", ".join(names_sorted)
        return f"The keys to {temple_name} can be found in {front} and {last}."


def create_message_for_hint(hint: Hint,
                            all_patches: Dict[int, GamePatches],
                            players_config: PlayersConfiguration,
                            hint_name_creator: LocationHintCreator,
                            location_formatters: Dict[HintLocationPrecision, LocationFormatter],
                            world_list: WorldList,
                            ) -> str:
    if hint.hint_type == HintType.JOKE:
        return color_text(TextColor.JOKE, hint_name_creator.create_joke_hint())

    elif hint.hint_type == HintType.RED_TEMPLE_KEY_SET:
        return create_temple_key_hint(all_patches, players_config.player_index, hint.dark_temple, world_list)

    else:
        assert hint.hint_type == HintType.LOCATION
        patches = all_patches[players_config.player_index]
        pickup_target = patches.pickup_assignment.get(hint.target)
        determiner, pickup_name = _calculate_pickup_hint(patches.pickup_assignment,
                                                         world_list,
                                                         hint.precision.item,
                                                         pickup_target,
                                                         players_config)

        use_title_formatting = True
        if hint.precision.include_owner and len(players_config.player_names) > 1:
            target_player = pickup_target.player if pickup_target is not None else players_config.player_index
            determiner = f"{players_config.player_names[target_player]}'s "
            use_title_formatting = False

        return location_formatters[hint.precision.location].format(
            Determiner(determiner, use_title_formatting),
            color_text(TextColor.ITEM, pickup_name),
            hint,
        )


def create_hints(all_patches: Dict[int, GamePatches],
                 players_config: PlayersConfiguration,
                 world_list: WorldList,
                 rng: Random,
                 ) -> list:
    """
    Creates the string patches entries that changes the Lore scans in the game for item pickups
    :param all_patches:
    :param players_config:
    :param world_list:
    :param rng:
    :return:
    """

    hint_name_creator = LocationHintCreator(world_list, rng, _JOKE_HINTS)
    patches = all_patches[players_config.player_index]

    location_formatters: Dict[HintLocationPrecision, LocationFormatter] = {
        HintLocationPrecision.KEYBEARER: TemplatedFormatter(
            "The Flying Ing Cache in {node} contains {determiner}{pickup}.", hint_name_creator),
        HintLocationPrecision.GUARDIAN: GuardianFormatter(),
        HintLocationPrecision.LIGHT_SUIT_LOCATION: TemplatedFormatter(
            "U-Mos's reward for returning the Sanctuary energy is {determiner}{pickup}.", hint_name_creator),
        HintLocationPrecision.DETAILED: TemplatedFormatter("{determiner.title}{pickup} can be found in {node}.",
                                                           hint_name_creator),
        HintLocationPrecision.WORLD_ONLY: TemplatedFormatter("{determiner.title}{pickup} can be found in {node}.",
                                                             hint_name_creator),
        HintLocationPrecision.RELATIVE_TO_AREA: RelativeAreaFormatter(world_list, patches),
        HintLocationPrecision.RELATIVE_TO_INDEX: RelativeItemFormatter(world_list, patches, players_config),
    }

    hints_for_asset: Dict[int, str] = {}
    for asset, hint in patches.hints.items():
        hints_for_asset[asset.asset_id] = create_message_for_hint(hint, all_patches, players_config, hint_name_creator,
                                                                  location_formatters, world_list)

    return [
        create_simple_logbook_hint(
            logbook_node.string_asset_id,
            hints_for_asset.get(logbook_node.string_asset_id, "Someone forgot to leave a message."),
        )
        for logbook_node in world_list.all_nodes
        if isinstance(logbook_node, LogbookNode)
    ]


def hide_hints(world_list: WorldList) -> list:
    """
    Creates the string patches entries that changes the Lore scans in the game
    completely useless text.
    :return:
    """

    return [create_simple_logbook_hint(logbook_node.string_asset_id, "Some item was placed somewhere.")
            for logbook_node in world_list.all_nodes if isinstance(logbook_node, LogbookNode)]



from typing import Dict

from randovania.game_description.game_patches import GamePatches
from randovania.game_description.resources import resource_info
from randovania.game_description.world_list import WorldList
from randovania.games.prime import echoes_items
from randovania.games.prime.patcher_file_lib.hint_name_creator import LocationHintCreator, create_simple_logbook_hint, \
    color_text, TextColor
from randovania.interface_common.players_configuration import PlayersConfiguration

_SKY_TEMPLE_KEY_SCAN_ASSETS = [
    0xD97685FE,
    0x32413EFD,
    0xDD8355C3,
    0x3F5F4EBA,
    0xD09D2584,
    0x3BAA9E87,
    0xD468F5B9,
    0x2563AE34,
    0xCAA1C50A,
]


def _sky_temple_key_name(key_number: int) -> str:
    return color_text(TextColor.ITEM, f"Sky Temple Key {key_number}")


def create_hints(all_patches: Dict[int, GamePatches],
                 players_config: PlayersConfiguration,
                 world_list: WorldList,
                 hide_area: bool,
                 ) -> list:
    """
    Creates the string patches entries that changes the Sky Temple Gateway hint scans with hints for where
    the STK actually are.
    :param all_patches:
    :param players_config:
    :param world_list:
    :param hide_area: Should the hint include only the world?
    :return:
    """
    location_hint_creator = LocationHintCreator(world_list, None, None)
    sky_temple_key_hints = {}

    for other_player, patches in all_patches.items():
        for pickup_index, target in patches.pickup_assignment.items():
            if target.player != players_config.player_index:
                continue
            resources = resource_info.convert_resource_gain_to_current_resources(target.pickup.resource_gain({}))

            for resource, quantity in resources.items():
                if quantity < 1:
                    continue

                try:
                    key_number = echoes_items.SKY_TEMPLE_KEY_ITEMS.index(resource.index) + 1
                except ValueError:
                    continue

                assert resource.index not in sky_temple_key_hints

                player_name = f"{players_config.player_names[other_player]}'s "

                sky_temple_key_hints[resource.index] = "{} is located in {}{}.".format(
                    _sky_temple_key_name(key_number),
                    color_text(TextColor.LOCATION, player_name) if len(all_patches) > 1 else "",
                    color_text(TextColor.LOCATION,
                               location_hint_creator.index_node_name(pickup_index, hide_area),
                               ),
                )

    for starting_resource, quantity in all_patches[players_config.player_index].starting_items.items():
        if quantity < 1:
            continue
        try:
            key_number = echoes_items.SKY_TEMPLE_KEY_ITEMS.index(starting_resource.index) + 1
        except ValueError:
            continue

        assert starting_resource.index not in sky_temple_key_hints
        sky_temple_key_hints[starting_resource.index] = "{} has no need to be located.".format(
            _sky_temple_key_name(key_number),
        )

    if len(sky_temple_key_hints) != len(echoes_items.SKY_TEMPLE_KEY_ITEMS):
        raise ValueError(
            "Expected to find {} Sky Temple Keys between pickup placement and starting items, found {}".format(
                len(echoes_items.SKY_TEMPLE_KEY_ITEMS),
                len(sky_temple_key_hints)
            ))

    return [
        create_simple_logbook_hint(_SKY_TEMPLE_KEY_SCAN_ASSETS[key_number], sky_temple_key_hints[key_index])
        for key_number, key_index in enumerate(echoes_items.SKY_TEMPLE_KEY_ITEMS)
    ]


def hide_hints() -> list:
    """
    Creates the string patches entries that changes the Sky Temple Gateway hint scans with hints for
    completely useless text.
    :return:
    """

    return [
        create_simple_logbook_hint(
            _SKY_TEMPLE_KEY_SCAN_ASSETS[key_number],
            "{} is lost somewhere in Aether.".format(_sky_temple_key_name(key_number + 1)),
        )
        for key_number in range(9)
    ]



import dataclasses
from typing import Tuple, List

from randovania.game_description.game_description import GameDescription
from randovania.game_description.game_patches import GamePatches
from randovania.game_description.node import EventNode, PickupNode, ResourceNode, Node
from randovania.game_description.requirements import ResourceRequirement, Requirement
from randovania.game_description.resources.resource_info import ResourceGain, CurrentResources, ResourceInfo


@dataclasses.dataclass(frozen=True)
class EventPickupNode(ResourceNode):
    event_node: EventNode
    pickup_node: PickupNode

    def __repr__(self):
        return "EventPickupNode({!r} -> {}+{})".format(
            self.name,
            self.event_node.resource().long_name,
            self.pickup_node.pickup_index.index,
        )

    @property
    def is_resource_node(self) -> bool:
        return True

    def resource(self) -> ResourceInfo:
        return self.pickup_node.pickup_index

    def requirement_to_leave(self, patches: GamePatches, current_resources: CurrentResources) -> Requirement:
        return ResourceRequirement(self.pickup_node.pickup_index, 1, False)

    def can_collect(self, patches, current_resources: CurrentResources, all_nodes: Tuple[Node, ...]) -> bool:
        event_collect = self.event_node.can_collect(patches, current_resources, all_nodes)
        pickup_collect = self.pickup_node.can_collect(patches, current_resources, all_nodes)
        return event_collect or pickup_collect

    def resource_gain_on_collect(self, patches, current_resources: CurrentResources, all_nodes: Tuple[Node, ...]
                                 ) -> ResourceGain:
        yield from self.event_node.resource_gain_on_collect(patches, current_resources, all_nodes)
        yield from self.pickup_node.resource_gain_on_collect(patches, current_resources, all_nodes)


def replace_with_event_pickups(game: GameDescription):
    for area in game.world_list.all_areas:
        nodes_to_replace: List[Tuple[EventNode, PickupNode]] = []

        for event_node in area.nodes:
            if not isinstance(event_node, EventNode):
                continue

            if len(area.connections[event_node]) != 1:
                continue

            next_node = list(area.connections[event_node].keys())[0]
            if not isinstance(next_node, PickupNode):
                continue

            if sum(1 for connections in area.connections.values() if next_node in connections) > 1:
                continue

            nodes_to_replace.append((event_node, next_node))

        for event_node, next_node in nodes_to_replace:
            combined_node = EventPickupNode(
                "EventPickup - {} + {}".format(event_node.event.long_name, next_node.name),
                event_node.heal or next_node.heal,
                next_node.location,
                event_node.index,
                event_node, next_node)

            # If the default node index is beyond one of the removed nodes, then fix it
            if area.default_node_index >= min(area.nodes.index(event_node), area.nodes.index(next_node)):
                object.__setattr__(area, "default_node_index", area.default_node_index - 1)

            area.nodes[area.nodes.index(event_node)] = combined_node
            game.world_list.add_new_node(area, combined_node)
            area.nodes.remove(next_node)

            for connections in area.connections.values():
                if event_node in connections:
                    connections[combined_node] = connections.pop(event_node)

            area.connections[combined_node] = area.connections.pop(next_node)



import io
import json
from pathlib import Path

import pytest
from PySide2.QtWidgets import QDialog
from mock import AsyncMock, patch, ANY

import randovania.game_description.pretty_print
from randovania.game_description import data_reader, data_writer
from randovania.game_description.area_location import AreaLocation
from randovania.game_description.requirements import Requirement
from randovania.gui.data_editor import DataEditorWindow


def test_apply_edit_connections_change(echoes_game_data,
                                       skip_qtbot,
                                       ):
    # Setup
    window = DataEditorWindow(echoes_game_data, None, False, True)
    skip_qtbot.addWidget(window)
    game = window.game_description

    landing_site = game.world_list.area_by_area_location(AreaLocation(1006255871, 1655756413))
    source = landing_site.node_with_name("Save Station")
    target = landing_site.node_with_name("Door to Service Access")

    # Run
    window.world_selector_box.setCurrentIndex(window.world_selector_box.findText("Temple Grounds (Sky Temple Grounds)"))
    window.area_selector_box.setCurrentIndex(window.area_selector_box.findText(landing_site.name))
    window._apply_edit_connections(source, target, Requirement.trivial())

    # Assert
    assert landing_site.connections[source][target] == Requirement.trivial()


def test_select_area_by_name(echoes_game_data,
                             skip_qtbot,
                             ):
    # Setup
    window = DataEditorWindow(echoes_game_data, None, False, True)
    skip_qtbot.addWidget(window)

    # Run
    window.focus_on_world("Torvus Bog")

    assert window.current_area.name != "Forgotten Bridge"
    window.focus_on_area("Forgotten Bridge")

    # Assert
    assert window.current_area.name == "Forgotten Bridge"


@pytest.mark.asyncio
@pytest.mark.parametrize("accept", [False, True])
@patch("randovania.gui.data_editor.ConnectionsEditor")
@patch("randovania.gui.data_editor.DataEditorWindow._apply_edit_connections", autospec=True)
async def test_open_edit_connection(mock_apply_edit_connections,
                                    mock_connections_editor,
                                    accept: bool,
                                    echoes_game_data,
                                    skip_qtbot,
                                    mocker,
                                    ):
    # Setup
    execute_dialog = mocker.patch("randovania.gui.lib.async_dialog.execute_dialog", new_callable=AsyncMock)
    window = DataEditorWindow(echoes_game_data, None, False, True)
    skip_qtbot.addWidget(window)

    editor = mock_connections_editor.return_value
    execute_dialog.return_value = QDialog.Accepted if accept else QDialog.Rejected

    # Run
    await window._open_edit_connection()

    # Assert
    mock_connections_editor.assert_called_once_with(window, window.resource_database, ANY)
    if accept:
        mock_apply_edit_connections.assert_called_once_with(window,
                                                            window.current_node,
                                                            window.current_connection_node,
                                                            editor.final_requirement)
    else:
        mock_apply_edit_connections.assert_not_called()


def test_create_node_and_save(tmpdir,
                              echoes_game_data,
                              skip_qtbot):
    # Setup
    db_path = Path(tmpdir.join("game.json"))

    window = DataEditorWindow(echoes_game_data, db_path, True, True)
    skip_qtbot.addWidget(window)

    # Run
    window._do_create_node("Some Node")
    window._save_as_internal_database()

    # Assert
    with db_path.open() as data_file:
        exported_data = json.load(data_file)
    exported_game = data_reader.decode_data(exported_data)

    output = io.StringIO()
    randovania.game_description.pretty_print.write_human_readable_world_list(exported_game, output)

    assert Path(tmpdir.join("game.txt")).read_text("utf-8") == output.getvalue()



from pathlib import Path

import pytest

import randovania


def test_get_configuration_default_missing(tmpdir, mocker):
    # Setup
    randovania.CONFIGURATION_FILE_PATH = None
    mocker.patch("randovania._get_default_configuration_path", return_value=Path(tmpdir).joinpath("missing.json"))

    # Run
    config = randovania.get_configuration()

    # Assert
    assert config == {
        "server_address": "http://127.0.0.1:5000",
        "socketio_path": "/socket.io",
    }


@pytest.mark.parametrize("from_configured", [False, True])
def test_get_configuration_file(tmpdir, mocker, from_configured):
    # Setup
    mocker.patch("randovania._get_default_configuration_path", return_value=Path(tmpdir).joinpath("provided.json"))
    if from_configured:
        Path(tmpdir).joinpath("configuration.json").write_text('{"foo": 5}')
        randovania.CONFIGURATION_FILE_PATH = Path(tmpdir).joinpath("configuration.json")
    else:
        Path(tmpdir).joinpath("provided.json").write_text('{"foo": 5}')

    # Run
    config = randovania.get_configuration()

    # Assert
    assert config == {
        "foo": 5,
    }


def test_get_invalid_configuration_file(tmpdir):
    # Setup
    randovania.CONFIGURATION_FILE_PATH = Path(tmpdir).joinpath("configuration.json")

    # Run
    with pytest.raises(FileNotFoundError):
        randovania.get_configuration()
